/* -*- mode: c++; indent-tabs-mode: nil -*- */
/** @file SFTPClient.qpp defines the SFTPClient class */
/*
  SFTPClient.qpp

  libssh2 SFTP client integration into qore

  Copyright 2009 Wolfgang Ritzinger
  Copyright 2010 - 2013 Qore Technologies, sro

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "SFTPClient.h"
#include "QC_SSH2Base.h"

static QoreHashNode *attr2hash(const LIBSSH2_SFTP_ATTRIBUTES &attr) {
   QoreHashNode *ret = new QoreHashNode;

   if (attr.flags & LIBSSH2_SFTP_ATTR_SIZE)
      ret->setKeyValue("size", new QoreBigIntNode(attr.filesize), 0);
   if (attr.flags & LIBSSH2_SFTP_ATTR_ACMODTIME) {
      ret->setKeyValue("atime", DateTimeNode::makeAbsolute(currentTZ(), (int64)attr.atime), 0);
      ret->setKeyValue("mtime", DateTimeNode::makeAbsolute(currentTZ(), (int64)attr.mtime), 0);
   }
   if (attr.flags & LIBSSH2_SFTP_ATTR_UIDGID) {
      ret->setKeyValue("uid", new QoreBigIntNode(attr.uid), 0);
      ret->setKeyValue("gid", new QoreBigIntNode(attr.gid), 0);
   }
   if (attr.flags & LIBSSH2_SFTP_ATTR_PERMISSIONS) {
      ret->setKeyValue("mode", new QoreBigIntNode(attr.permissions), 0);
      ret->setKeyValue("permissions", new QoreStringNode(mode2str(attr.permissions)), 0);
   }
  
   return ret;
}
/* Qore class Qore::SSH2::SFTPClient */

qore_classid_t CID_SFTPCLIENT;
QoreClass* QC_SFTPCLIENT;

// string SFTPClient::chdir(string path, timeout timeout = 60s){}
static AbstractQoreNode* SFTPClient_chdir_VsVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* path = HARD_QORE_STRING(args, 0);
    int64 timeout = HARD_QORE_INT(args, 1);
# 382 "QC_SFTPClient.qpp"
    return myself->sftp_chdir(path->getBuffer(), (int)timeout, xsink);
}

// nothing SFTPClient::chmod(string path, int mode, timeout timeout = 60s){}
static AbstractQoreNode* SFTPClient_chmod_VsViVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* path = HARD_QORE_STRING(args, 0);
    int64 mode = HARD_QORE_INT(args, 1);
    int64 timeout = HARD_QORE_INT(args, 2);
# 249 "QC_SFTPClient.qpp"
    // check if mode is in range
  if (mode != (mode & (int64)SFTP_UGOMASK)) {
      xsink->raiseException("SFTPCLIENT-PARAMETER-ERROR", "mode setting is only possible for user, group and other (no sticky bits)");
      return 0;
   }

  myself->sftp_chmod(path->getBuffer(), (int)mode, (int)timeout, xsink);
    return 0;
}

// SFTPClient::constructor(string url, softint port = 0) {}
static void SFTPClient_constructor_Vsvi(QoreObject* self, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* url = HARD_QORE_STRING(args, 0);
    int64 port = HARD_QORE_INT(args, 1);
# 65 "QC_SFTPClient.qpp"
QoreURL qurl(url);

   if (!qurl.getHost()) {
      xsink->raiseException("SFTPCLIENT-PARAMETER-ERROR", "no hostname found in URL '%s'", url->getBuffer());
      return;
   }

   if (qurl.getProtocol() && strcasecmp("sftp", qurl.getProtocol()->getBuffer())) {
      xsink->raiseException("SFTPCLIENT-PARAMETER-ERROR", "URL given in the first argument to SFTPClient::constructor() specifies invalid protocol '%s' (expecting 'sftp')", qurl.getProtocol()->getBuffer());
      return;
   }

   // create private data object
   self->setPrivate(CID_SFTPCLIENT, new SFTPClient(qurl, (int)port));
}

// SFTPClient::copy() {}
static void SFTPClient_copy(QoreObject* self, QoreObject* old, SFTPClient* myself, ExceptionSink* xsink) {
# 85 "QC_SFTPClient.qpp"
xsink->raiseException("SFTPCLIENT-COPY-ERROR", "copying sftp connection objects is not allowed");
}

// binary SFTPClient::getFile(string path, timeout timeout = 60s){}
static AbstractQoreNode* SFTPClient_getFile_VsVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* path = HARD_QORE_STRING(args, 0);
    int64 timeout = HARD_QORE_INT(args, 1);
# 271 "QC_SFTPClient.qpp"
    return myself->sftp_getFile(path->getBuffer(), (int)timeout, xsink);
}

// string SFTPClient::getTextFile(string path, timeout timeout = 60s, *string encoding){}
static AbstractQoreNode* SFTPClient_getTextFile_VsVtNs(QoreObject* self, SFTPClient* myself, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* path = HARD_QORE_STRING(args, 0);
    int64 timeout = HARD_QORE_INT(args, 1);
    const QoreStringNode* encoding = reinterpret_cast<const QoreStringNode*>(get_param(args, 2));
# 288 "QC_SFTPClient.qpp"
    const QoreEncoding *qe = encoding ? QEM.findCreate(encoding) : QCS_DEFAULT;
   return myself->sftp_getTextFile(path->getBuffer(), (int)timeout, qe, xsink);
}

// hash SFTPClient::info(){}
static AbstractQoreNode* SFTPClient_info(QoreObject* self, SFTPClient* myself, const QoreListNode* args, ExceptionSink* xsink) {
# 107 "QC_SFTPClient.qpp"
    return myself->sftp_info();
}

// hash SFTPClient::list(*string path, timeout timeout = 60s){}
static AbstractQoreNode* SFTPClient_list_NsVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* path = reinterpret_cast<const QoreStringNode*>(get_param(args, 0));
    int64 timeout = HARD_QORE_INT(args, 1);
# 141 "QC_SFTPClient.qpp"
    return myself->sftp_list(path ? path->getBuffer() : 0, (int)timeout, xsink);
}

// list SFTPClient::listFull(*string path, timeout timeout = 60s){}
static AbstractQoreNode* SFTPClient_listFull_NsVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* path = reinterpret_cast<const QoreStringNode*>(get_param(args, 0));
    int64 timeout = HARD_QORE_INT(args, 1);
# 170 "QC_SFTPClient.qpp"
    return myself->sftp_list_full(path ? path->getBuffer() : 0, (int)timeout, xsink);
}

// nothing SFTPClient::mkdir(string path, int mode = 0755, timeout timeout = 60s){}
static AbstractQoreNode* SFTPClient_mkdir_VsViVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* path = HARD_QORE_STRING(args, 0);
    int64 mode = HARD_QORE_INT(args, 1);
    int64 timeout = HARD_QORE_INT(args, 2);
# 349 "QC_SFTPClient.qpp"
    myself->sftp_mkdir(path->getBuffer(), (int)mode, (int)timeout, xsink);
    return 0;
}

// *string SFTPClient::path(){}
static AbstractQoreNode* SFTPClient_path(QoreObject* self, SFTPClient* myself, const QoreListNode* args, ExceptionSink* xsink) {
# 117 "QC_SFTPClient.qpp"
    return myself->sftp_path();
}

// int SFTPClient::putFile(binary bin, string path, int mode = 0644, timeout timeout = 60s){}
static int64 SFTPClient_putFile_VyVsViVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, ExceptionSink* xsink) {
    const BinaryNode* bin = HARD_QORE_BINARY(args, 0);
    const QoreStringNode* path = HARD_QORE_STRING(args, 1);
    int64 mode = HARD_QORE_INT(args, 2);
    int64 timeout = HARD_QORE_INT(args, 3);
# 309 "QC_SFTPClient.qpp"
    // transfer the file
   return myself->sftp_putFile((const char *)bin->getPtr(), bin->size(), path->getBuffer(), (int)mode, (int)timeout, xsink);
}

// int SFTPClient::putFile(string data, string path, int mode = 0644, timeout timeout = 60s){}
static int64 SFTPClient_putFile_VsVsViVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* data = HARD_QORE_STRING(args, 0);
    const QoreStringNode* path = HARD_QORE_STRING(args, 1);
    int64 mode = HARD_QORE_INT(args, 2);
    int64 timeout = HARD_QORE_INT(args, 3);
# 330 "QC_SFTPClient.qpp"
    // transfer the file
   return myself->sftp_putFile(data->getBuffer(), data->strlen(), path->getBuffer(), (int)mode, (int)timeout, xsink);
}

// nothing SFTPClient::removeFile(string path, timeout timeout = 60s){}
static AbstractQoreNode* SFTPClient_removeFile_VsVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* path = HARD_QORE_STRING(args, 0);
    int64 timeout = HARD_QORE_INT(args, 1);
# 213 "QC_SFTPClient.qpp"
    myself->sftp_unlink(path->getBuffer(), (int)timeout, xsink);
    return 0;
}

// nothing SFTPClient::rename(string old_name, string new_name, timeout timeout = 60s){}
static AbstractQoreNode* SFTPClient_rename_VsVsVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* old_name = HARD_QORE_STRING(args, 0);
    const QoreStringNode* new_name = HARD_QORE_STRING(args, 1);
    int64 timeout = HARD_QORE_INT(args, 2);
# 231 "QC_SFTPClient.qpp"
    myself->sftp_rename(old_name->getBuffer(), new_name->getBuffer(), (int)timeout, xsink);
    return 0;
}

// nothing SFTPClient::rmdir(string path, timeout timeout = 60s){}
static AbstractQoreNode* SFTPClient_rmdir_VsVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* path = HARD_QORE_STRING(args, 0);
    int64 timeout = HARD_QORE_INT(args, 1);
# 366 "QC_SFTPClient.qpp"
    myself->sftp_rmdir(path->getBuffer(), (int)timeout, xsink);
    return 0;
}

// *hash SFTPClient::stat(string path, timeout timeout = 60s){}
static AbstractQoreNode* SFTPClient_stat_VsVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* path = HARD_QORE_STRING(args, 0);
    int64 timeout = HARD_QORE_INT(args, 1);
# 195 "QC_SFTPClient.qpp"
    LIBSSH2_SFTP_ATTRIBUTES attr;
   int rc = myself->sftp_getAttributes(path->getBuffer(), &attr, (int)timeout, xsink);
   return rc < 0 ? 0 : attr2hash(attr);
}

DLLLOCAL void preinitSFTPClientClass() {
    QC_SFTPCLIENT = new QoreClass("SFTPClient", QDOM_NETWORK);
    CID_SFTPCLIENT = QC_SFTPCLIENT->getID();
    QC_SFTPCLIENT->setSystem();
}

DLLLOCAL QoreClass* initSFTPClientClass(QoreNamespace& ns) {
    if (!QC_SFTPCLIENT)
        preinitSFTPClientClass();

    // set parent class
    assert(QC_SSH2BASE);
    QC_SFTPCLIENT->addBuiltinVirtualBaseClass(QC_SSH2BASE);

    // string SFTPClient::chdir(string path, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethodExtended3("chdir", (q_method_t)SFTPClient_chdir_VsVt, false, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, stringTypeInfo, NULL, "path", timeoutTypeInfo, DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0), "timeout");

    // nothing SFTPClient::chmod(string path, int mode, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethodExtended3("chmod", (q_method_t)SFTPClient_chmod_VsViVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 3, stringTypeInfo, NULL, "path", bigIntTypeInfo, NULL, "mode", timeoutTypeInfo, DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0), "timeout");

    // SFTPClient::constructor(string url, softint port = 0) {}
    QC_SFTPCLIENT->setConstructorExtended3(SFTPClient_constructor_Vsvi, false, QC_NO_FLAGS, QDOM_DEFAULT, 2, stringTypeInfo, NULL, "url", softBigIntTypeInfo, zero(), "port");

    // SFTPClient::copy() {}
    QC_SFTPCLIENT->setCopy((q_copy_t)SFTPClient_copy);

    // binary SFTPClient::getFile(string path, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethodExtended3("getFile", (q_method_t)SFTPClient_getFile_VsVt, false, QC_NO_FLAGS, QDOM_DEFAULT, binaryTypeInfo, 2, stringTypeInfo, NULL, "path", timeoutTypeInfo, DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0), "timeout");

    // string SFTPClient::getTextFile(string path, timeout timeout = 60s, *string encoding){}
    QC_SFTPCLIENT->addMethodExtended3("getTextFile", (q_method_t)SFTPClient_getTextFile_VsVtNs, false, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 3, stringTypeInfo, NULL, "path", timeoutTypeInfo, DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0), "timeout", stringOrNothingTypeInfo, NULL, "encoding");

    // hash SFTPClient::info(){}
    QC_SFTPCLIENT->addMethodExtended3("info", (q_method_t)SFTPClient_info, false, QC_CONSTANT, QDOM_DEFAULT, hashTypeInfo);

    // hash SFTPClient::list(*string path, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethodExtended3("list", (q_method_t)SFTPClient_list_NsVt, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, hashTypeInfo, 2, stringOrNothingTypeInfo, NULL, "path", timeoutTypeInfo, DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0), "timeout");

    // list SFTPClient::listFull(*string path, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethodExtended3("listFull", (q_method_t)SFTPClient_listFull_NsVt, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, listTypeInfo, 2, stringOrNothingTypeInfo, NULL, "path", timeoutTypeInfo, DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0), "timeout");

    // nothing SFTPClient::mkdir(string path, int mode = 0755, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethodExtended3("mkdir", (q_method_t)SFTPClient_mkdir_VsViVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 3, stringTypeInfo, NULL, "path", bigIntTypeInfo, new QoreBigIntNode(0755), "mode", timeoutTypeInfo, DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0), "timeout");

    // *string SFTPClient::path(){}
    QC_SFTPCLIENT->addMethodExtended3("path", (q_method_t)SFTPClient_path, false, QC_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

    // int SFTPClient::putFile(binary bin, string path, int mode = 0644, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethodExtended3("putFile", (q_method_int64_t)SFTPClient_putFile_VyVsViVt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 4, binaryTypeInfo, NULL, "bin", stringTypeInfo, NULL, "path", bigIntTypeInfo, new QoreBigIntNode(0644), "mode", timeoutTypeInfo, DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0), "timeout");

    // int SFTPClient::putFile(string data, string path, int mode = 0644, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethodExtended3("putFile", (q_method_int64_t)SFTPClient_putFile_VsVsViVt, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 4, stringTypeInfo, NULL, "data", stringTypeInfo, NULL, "path", bigIntTypeInfo, new QoreBigIntNode(0644), "mode", timeoutTypeInfo, DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0), "timeout");

    // nothing SFTPClient::removeFile(string path, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethodExtended3("removeFile", (q_method_t)SFTPClient_removeFile_VsVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, stringTypeInfo, NULL, "path", timeoutTypeInfo, DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0), "timeout");

    // nothing SFTPClient::rename(string old_name, string new_name, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethodExtended3("rename", (q_method_t)SFTPClient_rename_VsVsVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 3, stringTypeInfo, NULL, "old_name", stringTypeInfo, NULL, "new_name", timeoutTypeInfo, DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0), "timeout");

    // nothing SFTPClient::rmdir(string path, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethodExtended3("rmdir", (q_method_t)SFTPClient_rmdir_VsVt, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, stringTypeInfo, NULL, "path", timeoutTypeInfo, DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0), "timeout");

    // *hash SFTPClient::stat(string path, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethodExtended3("stat", (q_method_t)SFTPClient_stat_VsVt, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, hashOrNothingTypeInfo, 2, stringTypeInfo, NULL, "path", timeoutTypeInfo, DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0), "timeout");

    return QC_SFTPCLIENT;
}
