/* -*- mode: c++; indent-tabs-mode: nil -*- */
/** @file SSH2Client.qpp defines the SSH2Client class */
/*
  QC_SSH2Client.qpp

  libssh2 ssh2 client integration into qore

  Copyright 2009 Wolfgang Ritzinger
  Copyright 2010 - 2013 Qore Technologies, sro

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "SSH2Client.h"

extern QoreClass* QC_SSH2BASE;
extern QoreClass* QC_SSH2CHANNEL;
/* Qore class Qore::SSH2::SSH2Client */

qore_classid_t CID_SSH2CLIENT;
QoreClass* QC_SSH2CLIENT;

// SSH2Client::constructor(string url, softint port = 0) {}
static void SSH2Client_constructor_Vsvi(QoreObject* self, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* url = HARD_QORE_STRING(args, 0);
    int64 port = HARD_QORE_INT(args, 1);
# 46 "QC_SSH2Client.qpp"
QoreURL qurl(url);

   if (!qurl.getHost()) {
      xsink->raiseException("SSH2CLIENT-PARAMETER-ERROR", "no hostname found in URL '%s'", url->getBuffer());
      return;
   }

   if (qurl.getProtocol() && strcasecmp("ssh", qurl.getProtocol()->getBuffer()) && strcasecmp("ssh2", qurl.getProtocol()->getBuffer())) {
      xsink->raiseException("SSH2CLIENT-PARAMETER-ERROR", "URL given in the first argument to SSH2Client::constructor() specifies invalid protocol '%s' (expecting 'ssh' or 'ssh2')", qurl.getProtocol()->getBuffer());
      return;
   }

   // create private data object
   SSH2Client* mySSH2Client = new SSH2Client(qurl, (int)port);
   self->setPrivate(CID_SSH2CLIENT, mySSH2Client);
}

// SSH2Client::copy() {}
static void SSH2Client_copy(QoreObject* self, QoreObject* old, SSH2Client* c, ExceptionSink* xsink) {
# 67 "QC_SSH2Client.qpp"
xsink->raiseException("SSH2CLIENT-COPY-ERROR", "copying ssh2 connection objects is not allowed");
}

// hash SSH2Client::info(){}
static AbstractQoreNode* SSH2Client_info(QoreObject* self, SSH2Client* c, const QoreListNode* args, ExceptionSink* xsink) {
# 88 "QC_SSH2Client.qpp"
    return c->ssh_info();
}

// SSH2Channel SSH2Client::openDirectTcpipChannel(string host, softint port, string source_host = "127.0.0.1", softint source_port = 22, timeout timeout = 60s){}
static QoreObject* SSH2Client_openDirectTcpipChannel_VsviVsviVt(QoreObject* self, SSH2Client* c, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* host = HARD_QORE_STRING(args, 0);
    int64 port = HARD_QORE_INT(args, 1);
    const QoreStringNode* source_host = HARD_QORE_STRING(args, 2);
    int64 source_port = HARD_QORE_INT(args, 3);
    int64 timeout = HARD_QORE_INT(args, 4);
# 121 "QC_SSH2Client.qpp"
    static const char *SSH2CLIENT_OPENDIRECTTCPIPCHANNEL_ERR = "SSH2CLIENT-OPENDIRECTTCPIPCHANNEL-ERROR";
   
   if (!port) {
      xsink->raiseException(SSH2CLIENT_OPENDIRECTTCPIPCHANNEL_ERR, "port number for forwarded channel as second argument to SSH2Client::openDirectTcpipChannel() cannot be zero");
      return 0;
   }

   if (!source_port) {
      xsink->raiseException(SSH2CLIENT_OPENDIRECTTCPIPCHANNEL_ERR, "source port number as fourth argument to SSH2Client::openDirectTcpipChannel() cannot be zero");
      return 0;
   }

   return c->openDirectTcpipChannel(xsink, host->getBuffer(), port, source_host->getBuffer(), source_port, timeout);
}

// SSH2Channel SSH2Client::openSessionChannel(timeout timeout = 60s){}
static QoreObject* SSH2Client_openSessionChannel_Vt(QoreObject* self, SSH2Client* c, const QoreListNode* args, ExceptionSink* xsink) {
    int64 timeout = HARD_QORE_INT(args, 0);
# 102 "QC_SSH2Client.qpp"
    return c->openSessionChannel(xsink, timeout);
}

// SSH2Channel SSH2Client::scpGet(string path, timeout timeout = 60s, *reference statinfo){}
static QoreObject* SSH2Client_scpGet_VsVtNr(QoreObject* self, SSH2Client* c, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* path = HARD_QORE_STRING(args, 0);
    int64 timeout = HARD_QORE_INT(args, 1);
    const ReferenceNode* statinfo = reinterpret_cast<const ReferenceNode*>(get_param(args, 2));
# 158 "QC_SSH2Client.qpp"
    ReferenceHolder<QoreHashNode> ref(statinfo ? new QoreHashNode : 0, xsink);

   ReferenceHolder<QoreObject> o(c->scpGet(xsink, path->getBuffer(), timeout, *ref), xsink);
   if (o && statinfo) {
      QoreTypeSafeReferenceHelper rh(statinfo, xsink);
      // a deadlock exception occurred accessing the reference's value pointer
      if (!rh || rh.assign(ref.release()))
	 return 0;
   }
   return o.release();
}

// SSH2Channel SSH2Client::scpPut(string remote_path, softint size, softint mode = 0644, *date mtime, *date atime, timeout timeout = 60s){}
static QoreObject* SSH2Client_scpPut_VsviviNdNdVt(QoreObject* self, SSH2Client* c, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* remote_path = HARD_QORE_STRING(args, 0);
    int64 size = HARD_QORE_INT(args, 1);
    int64 mode = HARD_QORE_INT(args, 2);
    const DateTimeNode* mtime = reinterpret_cast<const DateTimeNode*>(get_param(args, 3));
    const DateTimeNode* atime = reinterpret_cast<const DateTimeNode*>(get_param(args, 4));
    int64 timeout = HARD_QORE_INT(args, 5);
# 186 "QC_SSH2Client.qpp"
    static const char *SSH2CLIENT_SCPPUT_ERR = "SSH2CLIENT-SCPPUT-ERROR";
   
   if (size <= 0) {
      xsink->raiseException(SSH2CLIENT_SCPPUT_ERR, "invalid file size as second argument to SSH2Client::scpPut() (got invalid size "QLLD")", size);
      return 0;
   }

   return c->scpPut(xsink, remote_path->getBuffer(), size, (int)mode, mtime ? mtime->getEpochSeconds() : 0, atime ? atime->getEpochSeconds() : 0, timeout);
}

DLLLOCAL void preinitSSH2ClientClass() {
    QC_SSH2CLIENT = new QoreClass("SSH2Client", QDOM_NETWORK);
    CID_SSH2CLIENT = QC_SSH2CLIENT->getID();
    QC_SSH2CLIENT->setSystem();
}

DLLLOCAL QoreClass* initSSH2ClientClass(QoreNamespace& ns) {
    if (!QC_SSH2CLIENT)
        preinitSSH2ClientClass();

    // set parent class
    assert(QC_SSH2BASE);
    QC_SSH2CLIENT->addBuiltinVirtualBaseClass(QC_SSH2BASE);

    // SSH2Client::constructor(string url, softint port = 0) {}
    QC_SSH2CLIENT->setConstructorExtended3(SSH2Client_constructor_Vsvi, false, QC_NO_FLAGS, QDOM_DEFAULT, 2, stringTypeInfo, NULL, "url", softBigIntTypeInfo, zero(), "port");

    // SSH2Client::copy() {}
    QC_SSH2CLIENT->setCopy((q_copy_t)SSH2Client_copy);

    // hash SSH2Client::info(){}
    QC_SSH2CLIENT->addMethodExtended3("info", (q_method_t)SSH2Client_info, false, QC_CONSTANT, QDOM_DEFAULT, hashTypeInfo);

    // SSH2Channel SSH2Client::openDirectTcpipChannel(string host, softint port, string source_host = "127.0.0.1", softint source_port = 22, timeout timeout = 60s){}
    QC_SSH2CLIENT->addMethodExtended3("openDirectTcpipChannel", (q_method_t)SSH2Client_openDirectTcpipChannel_VsviVsviVt, false, QC_NO_FLAGS, QDOM_DEFAULT, QC_SSH2CHANNEL->getTypeInfo(), 5, stringTypeInfo, NULL, "host", softBigIntTypeInfo, NULL, "port", stringTypeInfo, new QoreStringNode("127.0.0.1"), "source_host", softBigIntTypeInfo, new QoreBigIntNode(22), "source_port", timeoutTypeInfo, DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0), "timeout");

    // SSH2Channel SSH2Client::openSessionChannel(timeout timeout = 60s){}
    QC_SSH2CLIENT->addMethodExtended3("openSessionChannel", (q_method_t)SSH2Client_openSessionChannel_Vt, false, QC_NO_FLAGS, QDOM_DEFAULT, QC_SSH2CHANNEL->getTypeInfo(), 1, timeoutTypeInfo, DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0), "timeout");

    // SSH2Channel SSH2Client::scpGet(string path, timeout timeout = 60s, *reference statinfo){}
    QC_SSH2CLIENT->addMethodExtended3("scpGet", (q_method_t)SSH2Client_scpGet_VsVtNr, false, QC_NO_FLAGS, QDOM_DEFAULT, QC_SSH2CHANNEL->getTypeInfo(), 3, stringTypeInfo, NULL, "path", timeoutTypeInfo, DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0), "timeout", referenceOrNothingTypeInfo, NULL, "statinfo");

    // SSH2Channel SSH2Client::scpPut(string remote_path, softint size, softint mode = 0644, *date mtime, *date atime, timeout timeout = 60s){}
    QC_SSH2CLIENT->addMethodExtended3("scpPut", (q_method_t)SSH2Client_scpPut_VsviviNdNdVt, false, QC_NO_FLAGS, QDOM_DEFAULT, QC_SSH2CHANNEL->getTypeInfo(), 6, stringTypeInfo, NULL, "remote_path", softBigIntTypeInfo, NULL, "size", softBigIntTypeInfo, new QoreBigIntNode(0644), "mode", dateOrNothingTypeInfo, NULL, "mtime", dateOrNothingTypeInfo, NULL, "atime", timeoutTypeInfo, DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0), "timeout");

    return QC_SSH2CLIENT;
}
